; An Emacs Lisp static analysis tool that does free name analysis across a number of different Emacs Lisp source files.
; The tool is also able to do systematic renamings of functions in a set of Emacs Lisp source files.

; Copyright (C) 2006  Kurt Normark, normark@cs.aau.dk.
; 
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
; 
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; ---------------------------------------------------------------------------------------------------------------
; Before this file is loaded, you must define the variable el-static-analysis-software-dir.
; This is typically done in your .emacs file, along the lines of the of the following comment:

; (defvar el-static-analysis-software-dir "..." 
;  "The directory that holds the Emacs Lisp static analysis tool. Ends with a forward slash character.")

; ---------------------------------------------------------------------------------------------------------------

(defconst elan-list-of-toplevel-defining-forms '(defun defvar defconst defcustom setq defmacro defsubst) 
  "A list of the names of forms considered as being top-level, and subject for analysis in this tool")

(defconst elan-forms-to-ignore '(load provide require global-set-key define-key laml-define-key define-derived-mode defgroup)
  "A list of forms ignored - and in which we do not care of free names." )

(defconst elan-form-problems nil "A global variable through which to report problem forms. That is in reality forms which are not analyzed by the tool.")


; ---------------------------------------------------------------------------------------------------------------
; CODE BASE DEFINITIONS

(defvar elan-current-analysis-code-base nil "A list of source files - full paths - which together constitute the current code base for analysis")

(defvar elan-current-definition-code-base nil "A list of source files - full paths - which together constitute the a large, surrounding code base of all known definitions. 
   This code base can be thought of as your Emacs lisp universe.")

(defvar elan-analysis-free-name-entry-list nil "A free name entry list of each analyzed form in the analysis code base.
  Thus, this list has an element for each top level definition being analyzed.
  Some forms, determined by the variable forms-to-ignore, are totally ignored by the analysis. 
  Among the non-ignored forms, the function defining-form-p selects those for analys.
  The free name entry reflects all the free names of the top-level definition, but without mentioning of built-in functions, byte-coded functions and aliases. 
  Notice that some of the free names may be defined in sibling definitions in the code base.
  A free name entry is a list of the form (definition-name kind-of-definition file-path . free-name-list)")

(defvar elan-definition-free-name-entry-list nil "The list of free name entries of the definition code base.
   Of minor importance.")

(defvar elan-defined-names-in-definition-codebase nil "The list of the currently defined names in the definition code base. Each entry in the list is a cons pair, due to its use for completion purposes.")

(defvar elan-defined-names-in-analysis-codebase nil "The list of the currently defined names in the analysis code base. Each entry is a cons pair, due to its use for completion purposes.")

(defvar elan-actual-free-name-list nil "A flat list of free names in the analysis code base. 
   This list does not include built-in functions, aliases, byte-code functions, or functions defined in the analysis code base.")

; ---------------------------------------------------------------------------------------------------------------
; Interactive commands:

(defun new-analysis-code-base (code-base-file-path)
  "Make an analysis code base. This is the set of Scheme source files whose definitions can be analysed for
free names. Takes as parameter either a cbs file which enumerates a number of source files, or a single el file.
A cbs file may contain a number of absolute paths to a number to emacs lisp files, 
a number of relative emacs lisp files - relative to the directory with the root code base file,  
or recursively, other code base files). This functions carries out the free name analysis, and reflect the analysis in the variables 
analysis-free-name-entry-list and defined-names-in-analysis-codebase."
  (interactive "fSelect analysis code base (cbs or el) file:  ")
  (let ((code-base-file-path-extension (elan-file-name-extension code-base-file-path))
        )
    (if (or (equal code-base-file-path-extension "el") (equal code-base-file-path-extension "EL"))
        (let ((code-base-file-list (list (elan-eliminate-tilde-prefix (expand-file-name code-base-file-path)))))
          (setq elan-current-analysis-code-base code-base-file-list)
          (reanalyse-analysis-codebase elan-current-analysis-code-base)
          (message (concat "DONE. " (int-to-string (length elan-analysis-free-name-entry-list)) " definitions in analysis codebase.")))

      (let ((code-base-file-list (elan-file-read-code-base code-base-file-path)))
        (setq elan-current-analysis-code-base code-base-file-list)
        (reanalyse-analysis-codebase elan-current-analysis-code-base)
        (message (concat "DONE. " (int-to-string (length elan-analysis-free-name-entry-list)) " definitions in analysis codebase."))))
    ))


(defun reanalyse-analysis-codebase (analysis-code-base)
  "Carry out the free name analysis of the files in the analysis code base.
Notice that the analysis is carried out on files, not buffers. This command does not save any file before analysis.
Re-assign the variables analysis-free-name-entry-list, defined-names-in-analysis-codebase, and actual-free-name-list."
  (interactive (list elan-current-analysis-code-base))
  (setq elan-analysis-free-name-entry-list (elan-free-names-of-all-forms-all-paths analysis-code-base))   ; Each entry analyzed in isolation.
  (setq elan-defined-names-in-analysis-codebase                                                      ; Just a listing of all definitions
       (mapcar (function (lambda (fne) (cons (elan-symbol-to-string (elan-defining-name-of-free-name-entry fne)) nil)))
               elan-analysis-free-name-entry-list))
  (let* ((fne-lst (elan-free-names-of-all-forms-outside-code-base analysis-code-base)))              ; A free name list, which only focuses on free name outside the analysis code base.
      (setq elan-actual-free-name-list 
            (elan-remove-duplicates 
             (elan-accumulate-right (function append) nil
                               (mapcar (function elan-free-name-list-of-free-name-entry) fne-lst)))))
  (message "Done. The analysis codebase has been analysed from the saved files.")
  )
  

(defun new-definition-code-base (code-base-file-path)
  "Make a definition code base. This is the set of definitions taken into account when code is
inserted into an existing source file with free names. Accepts either a cbs file or an emacs lisp file."
  (interactive "fSelect definition code base (cbs) file:  ")
  (let ((code-base-file-path-extension (elan-file-name-extension code-base-file-path))
        )
    (if (equal code-base-file-path-extension "el")
        (let* ((code-base-file-list (list (elan-eliminate-tilde-prefix (expand-file-name code-base-file-path))))
               (free-name-entry-list (elan-free-names-of-all-forms-all-paths code-base-file-list))  ; free names not actually not needed
              )
          (setq elan-current-definition-code-base code-base-file-list)
          (setq elan-definition-free-name-entry-list free-name-entry-list)
          (setq elan-defined-names-in-definition-codebase 
                (mapcar (function (lambda (fne) (cons (elan-symbol-to-string (elan-defining-name-of-free-name-entry fne)) nil)))
                        free-name-entry-list))

          (message (concat "DONE. " (int-to-string (length elan-defined-names-in-definition-codebase)) " definitions in definition code base.")))

      (let* ((code-base-file-list (elan-file-read-code-base code-base-file-path))
             (free-name-entry-list (elan-free-names-of-all-forms-all-paths code-base-file-list))   ; free names not actually not needed
            )
        (setq elan-current-definition-code-base code-base-file-list)
        (setq elan-definition-free-name-entry-list free-name-entry-list)
        (setq elan-defined-names-in-definition-codebase 
              (mapcar (function (lambda (fne) (cons (elan-symbol-to-string (elan-defining-name-of-free-name-entry fne)) nil)))
                      free-name-entry-list))
        (message (concat "DONE. " (int-to-string (length elan-defined-names-in-definition-codebase)) " definitions in definition codebase."))))))


(defun duplicates-in-current-code-base ()
  "Reported duplicated definitions in the current code base"
  (interactive)
  (let* ((buf-name "*DuplicateDefinitions*")
         (report-buffer (if (get-buffer buf-name) (get-buffer buf-name) (generate-new-buffer buf-name)))
         (duplicate-free-name-entries (elan-duplicate-definitions elan-analysis-free-name-entry-list))
        )
    (if (> (length duplicate-free-name-entries) 0)
        (progn
          (delete-other-windows)
          (split-window-vertically)
          (show-buffer (other-window 1) report-buffer)
          (set-buffer report-buffer)
          (erase-buffer)

          (insert "DUPLICATED DEFINITIONS - ANALYSIS RESULTS") (insert elan-CR) (insert elan-CR)

          (insert "Legend:")  (insert elan-CR)
          (insert "  <NAME>: [<KIND>]     <SOURCE-FILE-1>") (insert elan-CR)
          (insert "  <NAME>: [<KIND>]     <SOURCE-FILE-2>") (insert elan-CR)
          (insert "  <NAME> is defined in both <SOURCE-FILE-1>, <SOURCE-FILE-2>, etc.") (insert elan-CR) (insert elan-CR)


          (let ((dfn-lst duplicate-free-name-entries))
            (while (not (null dfn-lst))
              (insert (elan-symbol-to-string (elan-defining-name-of-free-name-entry (car dfn-lst)))) (insert ":   ")
              (insert "[") (insert (elan-symbol-to-string (elan-kind-of-free-name-entry (car dfn-lst)))) (insert "]") (insert "   ")
              (insert (elan-file-path-of-free-name-entry (car dfn-lst)))
              (insert elan-CR)
              (setq dfn-lst (cdr dfn-lst))))

          (goto-char (point-min))
          (other-window 1))

         (progn
           (beep)
           (message "No duplicates among the definitions in the analysis codebase.")) )))


(defun free-names-outside-current-code-base ()
  "Report of free names of all definitions outside the current code base. Only non-subrs, non-copiled free names
FROM OUTSIDE the actual files in the analysis code base are reported. Thus, we actually report on dependencies outside
the current analysis code base."
  (interactive)
  (let* ((fne-lst (elan-free-names-of-all-forms-outside-code-base elan-current-analysis-code-base))
        )
      (let* ((buf-name "*FreeNames*")
             (report-buffer (if (get-buffer buf-name) (get-buffer buf-name) (generate-new-buffer buf-name))))

        (setq elan-actual-free-name-list 
              (elan-remove-duplicates 
                (elan-accumulate-right (function append) nil (mapcar (function elan-free-name-list-of-free-name-entry) fne-lst))))

        (delete-other-windows)
        (split-window-vertically)
        (show-buffer (other-window 1) report-buffer)
        (set-buffer report-buffer)
        (erase-buffer)

        (insert "FREE NAME ANALYSIS RESULTS") (insert elan-CR) (insert elan-CR)

        (insert "Legend:")  (insert elan-CR)
        (insert "  <NAME>: [<KIND>, \"<SOURCE-FILE>\"]     <FREE-NAME-1>, <FREE-NAME-2>, ...") (insert elan-CR) 
        (insert "  The definition <NAME> of <KIND>, located in <SOURCE-FILE>, have") (insert elan-CR)
        (insert "   <FREE-NAME-1>, <FREE-NAME-2> outside the the analysis code base.") (insert elan-CR) (insert elan-CR)

        (while (not (null fne-lst))
          (let* ((free-name-lst (elan-free-name-list-of-free-name-entry (car fne-lst)))
                )
           (if (not (null free-name-lst))
            (let* ((full-file-path (elan-file-path-of-free-name-entry (car fne-lst)))
                   (file-name  (file-name-nondirectory full-file-path))
                  )
              (insert (elan-symbol-to-string (elan-defining-name-of-free-name-entry (car fne-lst)))) (insert ":  ")
              (insert "[") (insert (elan-symbol-to-string (elan-kind-of-free-name-entry (car fne-lst)))) (insert ",")
                           (insert " \"") (insert file-name) (insert "\"") 
               (insert "]") (insert "     ")
              (elan-insert-free-name-report free-name-lst)
              (insert elan-CR)))
            (setq fne-lst (cdr fne-lst))))

        (goto-char (point-min))

        (other-window 1)))
  (message "DONE")
 )

(defvar elan-previous-free-names-outside-file-dir nil)
(defvar elan-previous-free-names-outside-file-filename nil)

; Not used anymore. The analysis codebase can be a single file.
(defun free-names-outside-file (file-path-0)
  "Report of free names of all definitions in file-path. Only non-subrs, non-copiled free names
FROM OUTSIDE the actual file are reported. Thus, we actually report on dependencies outside
this file (the file in file-path)."

  ; Remember previous file name:
  (interactive  
    (list 
      (read-file-name "Which Emacs Lisp file: "
        (if elan-previous-free-names-outside-file-dir elan-previous-free-names-outside-file-dir (elan-current-directory))
        nil
        t
        (if elan-previous-free-names-outside-file-filename elan-previous-free-names-outside-file-filename nil))))

  (let* ((file-path (elan-eliminate-tilde-prefix (expand-file-name file-path-0)))
         (fne-lst (elan-free-names-of-all-forms-outside-file file-path))
        )
      (let* ((buf-name "*FreeNames*")
             (report-buffer (if (get-buffer buf-name) (get-buffer buf-name) (generate-new-buffer buf-name))))

        (setq elan-actual-free-name-list 
              (elan-remove-duplicates (elan-accumulate-right (function append) nil (mapcar (function elan-free-name-list-of-free-name-entry) fne-lst))))

        (delete-other-windows)
        (split-window-vertically)
        (show-buffer (other-window 1) report-buffer)
        (set-buffer report-buffer)
        (erase-buffer)

        (while (not (null fne-lst))
          (let ((free-name-lst (elan-free-name-list-of-free-name-entry (car fne-lst))))
           (if (not (null free-name-lst))
            (let*  ((full-file-path (elan-file-path-of-free-name-entry (car fne-lst)))
                    (file-name  (file-name-nondirectory full-file-path))
                    )
              (insert (elan-symbol-to-string (elan-defining-name-of-free-name-entry (car fne-lst)))) (insert ":  ")
              (insert "[") (insert (elan-symbol-to-string (elan-kind-of-free-name-entry (car fne-lst)))) (insert ",")
                           (insert " \"") (insert file-name) (insert "\"") 
               (insert "]") (insert "     ")
              (elan-insert-free-name-report free-name-lst)
              (insert elan-CR)))
            (setq fne-lst (cdr fne-lst))))

        (goto-char (point-min))

        (other-window 1)))
  (setq elan-previous-free-names-outside-file-dir (file-name-directory file-path-0))
  (setq elan-previous-free-names-outside-file-filename (file-name-nondirectory file-path-0))
  (message "DONE. The variable  actual-free-name-list  contains the appended list of free names involved.")
 )

(defun insert-remote-definition (name-str)
  "Insert the defintion of name at point. name corresponds to some definition
in the current code base."
  (interactive  
    (list (completing-read "Insert which definition: " 
                   elan-defined-names-in-definition-codebase nil nil)))
  (let* ((name (elan-string-to-symbol name-str))
         (name-fne-list (elan-filter (function (lambda (fne) (eq name (elan-defining-name-of-free-name-entry fne)))) elan-definition-free-name-entry-list))
        )
    (let* ((fne (car name-fne-list))
           (name-fne (elan-kind-of-free-name-entry fne))
           (def (elan-find-el-definition name (elan-kind-of-free-name-entry fne) (elan-file-path-of-free-name-entry fne)))
          )
      (if def
          (insert def)
          (error (concat "Could not locate "  (elan-symbol-to-string name) " in " (elan-file-path-of-free-name-entry name-fne)))))

    (if (> (length name-fne-list) 1) 
        (progn
           (beep)
           (message (concat (elan-symbol-to-string name) " " "is ambiguous. Using the one from" " " (file-name-nondirectory (elan-file-path-of-free-name-entry (car name-fne-list)))))))))

(defvar elan-free-name-problems nil "The names of the definitions which could not be inserted by means of the command insert-definitions-of-actual-free-names")

(defun insert-definitions-of-actual-free-names ()
  "Relative to the current code bases, and relative to the lastest free name analysis
insert as many definitions of free names as possible into the current buffer. 
Insertion is done at the position of point (at the current position of the cursor). 
The free names are taken from the variable actual-free-name-list, as defined by the latest analysis of the analysis codebase.
Return the number of insertions actually done."
  (interactive)
  (let ((problem-list nil)
        (ok-list nil))
   (mapcar
    (function (lambda (name)
      (let ((name-fne-list 
               (elan-filter (function (lambda (fne) (eq name (elan-defining-name-of-free-name-entry fne)))) elan-definition-free-name-entry-list))
           )
        (if (not (null name-fne-list))
          (let* ((fne (car name-fne-list))
                 (name-fne (elan-kind-of-free-name-entry fne))
                 (def (elan-find-el-definition name (elan-kind-of-free-name-entry fne) (elan-file-path-of-free-name-entry fne)))
                 )
            (if def
                (progn 
                  (insert def) (insert elan-CR) (insert elan-CR)
                  (setq ok-list (cons name ok-list))
                )
                (setq problem-list (cons name problem-list))))
          (setq problem-list (cons name problem-list))))))

    elan-actual-free-name-list)

   (setq elan-free-name-problems problem-list)
   (message (concat "DONE. " 
             "Inserted " (int-to-string (length ok-list)) " definitions. "
             (if (> (length problem-list) 0)
                (concat 
                  (int-to-string (length problem-list))
                  " non-inserted definitions. See the variable free-name-problems.") "")))
   (length ok-list)
 )
)

(defun reanalyse-and-insert ()
  "Do the free name analysis of the code base again. The situation is the following: We have just inserted
a number of definition, on which the current analysis code base depends. This may introduce more free names (in the inserted
definitions) which also must be inserted. These extra definitions are inserted by this command."
  (interactive)

  (save-buffer)
  (let* ((fne-lst (elan-free-names-of-all-forms-outside-code-base elan-current-analysis-code-base))
        )
    (setq elan-actual-free-name-list 
              (elan-remove-duplicates 
                (elan-accumulate-right (function append) nil
                   (mapcar (function elan-free-name-list-of-free-name-entry) fne-lst))))

    (insert-definitions-of-actual-free-names)))

(defun repetitive-reanalyse-and-insert ()
  "Repetitive activation of reanalyse-and-insert as long as something is inserted."
  (interactive)

  ; First time
  (let ((inserted-number-of-definitions-1 (insert-definitions-of-actual-free-names))   ; First time her
        (actually-inserted-definitions 1))  ; just non-0 init value
    (if (> inserted-number-of-definitions-1 0)
        (progn
          (save-buffer)
          (while (> actually-inserted-definitions 0)
	    (let* ((fne-lst (elan-free-names-of-all-forms-outside-code-base elan-current-analysis-code-base))
		   )
	      (setq elan-actual-free-name-list 
		    (elan-remove-duplicates 
		     (elan-accumulate-right (function append) nil
				       (mapcar (function elan-free-name-list-of-free-name-entry) fne-lst))))

	      (let ((inserted-number-of-definitions (insert-definitions-of-actual-free-names)))
                (setq actually-inserted-definitions inserted-number-of-definitions))

              (save-buffer) ))))))
       

; Make a buffer with  name replacement list for the definitions in the analysis code base.
; Systematically prefix all definitions with prefix (a string).
; The content of the *NameReplacements* buffer will control a forthcoming renaming of free names
; in the analysis code base.
(defun new-name-replacements-of-analysis-codebase (prefix)
  (interactive "sPrefix: ")
  (let* ((buf-name "*NameReplacements*")
         (name-replacement-buffer (if (get-buffer buf-name) (get-buffer buf-name) (generate-new-buffer buf-name)))
         (names elan-defined-names-in-analysis-codebase)
         (interactive-clue "    ; INTERACTIVE")
        )

    (delete-other-windows)
    (split-window-vertically)
    (show-buffer (other-window 1) name-replacement-buffer)
    (set-buffer name-replacement-buffer)
    (erase-buffer)


    (insert "(") (insert elan-CR)
    (while (not (null names))
      (let* ((name (car (car names)))   ; extra car:   due to extra list level in defined-names-in-analysis-codebase.
             (name-replacement (concat prefix name))
             (is-interactive (commandp (elan-string-to-symbol name)))
            )
        (insert (concat "   " "(" name "  " (if is-interactive name name-replacement) ")" (if is-interactive interactive-clue "") elan-CR))
        (setq names (cdr names))))
    (insert ")") (insert elan-CR)

    (goto-char (point-min))
    (setq elan-name-replacement-list (elan-read-name-replacement-buffer))
    (goto-char (point-min))

    (other-window 1))

    (message "You can edit the replacement list. A fortcoming renaming is taken directly from the buffer.")

)

(defun elan-read-name-replacement-buffer ()
  (save-excursion
    (set-buffer (get-buffer "*NameReplacements*"))
    (goto-char (point-min))
    (read (get-buffer (current-buffer)))))


(defun rename-free-names-in-analysis-code-base ()
  "Read the name-replacement-list from the *NameReplacements* buffer and
carry out renamings in all files of the analysis code buffer. Does not rewrite (save) buffers to files."
  (interactive)
  (save-window-excursion
    (setq elan-name-replacement-list (elan-read-name-replacement-buffer))
    (let ((code-base-list elan-current-analysis-code-base))
      (while (not (null code-base-list))
        (find-file (car code-base-list))
        (elan-do-rename-all-forms-in-current-buffer)
        (goto-char (point-min))
        (setq code-base-list (cdr code-base-list))))
    (message "RENAMING DONE. The source files in affected buffers have NOT been saved!!!")))

(defun code-bases-info ()
  "Display information about both the analysis codebase and the definition codebase."
  (interactive)
  (delete-other-windows)
  (split-window-vertically)
  (other-window 1)
  (let ((buf (get-buffer-create "*codebase-info*")))
    (set-buffer buf)

    (toggle-read-only -1)
    (erase-buffer)

    (insert "INFORMATION ABOUT CURRENT CODEBASES") (insert elan-CR) (insert elan-CR)
   
    (insert (concat "Number of files in ANALYSIS codebase: " 
                     (int-to-string (length elan-current-analysis-code-base)))) (insert elan-CR)
    (insert (concat "Number of definitions in ANALYSIS codebase: ") 
    (int-to-string (length elan-defined-names-in-analysis-codebase))) (insert elan-CR)
    (insert (concat "Files in current ANALYSIS codebase: ")) (insert (concat elan-CR "   "))
    (insert (elan-list-to-string elan-current-analysis-code-base (concat elan-CR "   ")))  (insert elan-CR) (insert elan-CR) (insert elan-CR) 

    (insert (concat "Number of definitions in DEFINITION code base: " 
                    (int-to-string (length elan-defined-names-in-definition-codebase)))) (insert elan-CR) 
    (insert (concat "Number of files in DEFINITION codebase: " 
                     (int-to-string (length elan-current-definition-code-base)))) (insert elan-CR) 
    (insert (concat "Files in current definition codebase: ")) (insert (concat elan-CR "   "))
    (insert (elan-list-to-string elan-current-definition-code-base (concat elan-CR "   ")))  (insert elan-CR) (insert elan-CR)

    (toggle-read-only 1)
    (set-buffer-modified-p nil)
    
    (goto-char (point-min))
    (show-buffer (selected-window) buf)))


(defun el-analysis-help ()
  "Show help file of this tool."
  (interactive)
  (let* ((buf-name "*ElAnalysisHelp*")
         (help-buffer (if (get-buffer buf-name) (get-buffer buf-name) (generate-new-buffer buf-name))))
 
    (delete-other-windows)
    (split-window-vertically)
    (other-window 1)

    (show-buffer (other-window 1) help-buffer)
    (set-buffer help-buffer)
    (toggle-read-only -1)
    (erase-buffer)
    (insert-file (concat el-static-analysis-software-dir "help.txt"))
    (toggle-read-only 1)
    (set-buffer-modified-p nil)
    (other-window 1)))


; ---------------------------------------------------------------------------------------------------------------
; Menu support

(define-key global-map [menu-bar tools el-static-analysis]
     (cons "Emacs Lisp analysis" (make-sparse-keymap "Static Emacs Lisp analysis")))

(define-key global-map [menu-bar tools el-static-analysis el-analysis-help]
      '("Help" . el-analysis-help))

(define-key global-map [menu-bar tools el-static-analysis code-bases-info]
      '("Codebases info" . code-bases-info))

(define-key  global-map [menu-bar tools el-static-analysis separator-1]
    '("----"))

(define-key global-map [menu-bar tools el-static-analysis rename-free-names-in-analysis-code-base]
      '("Rename free names in analysis code base" . rename-free-names-in-analysis-code-base))

(define-key global-map [menu-bar tools el-static-analysis new-name-replacements-of-analysis-codebase]
      '("New name replacement list of analysis code base" . new-name-replacements-of-analysis-codebase))

(define-key  global-map [menu-bar tools el-static-analysis separator-3]
    '("----"))

(define-key global-map [menu-bar tools el-static-analysis insert-remote-definition]
      '("Insert remote Emacs Lisp definition..." . insert-remote-definition))

(define-key global-map [menu-bar tools el-static-analysis repetitive-reanalyse-and-insert]
     '("Reanalyse and insert definitions repetitive" . repetitive-reanalyse-and-insert))

(define-key global-map [menu-bar tools el-static-analysis reanalyse-and-insert]
     '("Reanalyse and insert definitions" . reanalyse-and-insert))

(define-key global-map [menu-bar tools el-static-analysis insert-definitions-of-actual-free-names]
      '("Insert definitions of free names" . insert-definitions-of-actual-free-names))

(define-key  global-map [menu-bar tools el-static-analysis separator-2]
    '("----"))

;(define-key global-map [menu-bar tools el-static-analysis free-names-outside-file]
;      '("Free names defined outside source file..." . free-names-outside-file))

(define-key global-map [menu-bar tools el-static-analysis duplicates-in-current-code-base]
      '("Show duplicate definitions in analysis codebase" . duplicates-in-current-code-base))

(define-key global-map [menu-bar tools el-static-analysis free-names-outside-current-code-base]
      '("Show free name analysis of analysis code base" . free-names-outside-current-code-base))

(define-key  global-map [menu-bar tools el-static-analysis separator-4]
    '("----"))

(define-key global-map [menu-bar tools el-static-analysis reanalyse-analysis-codebase]
      '("Reanalyse analysis codebase" . reanalyse-analysis-codebase))

(define-key  global-map [menu-bar tools el-static-analysis separator-0]
    '("----"))

(define-key global-map [menu-bar tools el-static-analysis new-analysis-code-base]
      '("New Codebase for analysis..." . new-analysis-code-base))

(define-key global-map [menu-bar tools el-static-analysis new-definition-code-base]
      '("New Codebase for definitions..." . new-definition-code-base))


; ---------------------------------------------------------------------------------------------------------------

(defun elan-free-names-of-all-forms-all-paths (file-path-list)
  "Return a list of free name entries of all files in file-path-list. Only interfile free names are included."
  (setq elan-form-problems nil)
  (elan-flatten (mapcar (function elan-free-names-of-all-forms) file-path-list)))

(defun elan-free-names-of-all-forms (file-path)
  "Return a list of free name entries of all definitions in file-path"
 (let* ((form-list (elan-file-read-all file-path))
        (form-list-non-ignore (elan-filter (lambda (form) (not (memq (car form) elan-forms-to-ignore))) form-list))
        (form-list-ok (elan-filter (function elan-defining-form-p) form-list-non-ignore))
        (form-list-problems (elan-filter (function (lambda (form) (not (elan-defining-form-p form)))) form-list-non-ignore))
       )
   (if (= 0 (length form-list-problems))
       (message "Done")
       (message (concat "Problems: " (int-to-string (length form-list-problems)) ". See variable form-problems")))

   (setq elan-form-problems (append form-list-problems elan-form-problems))
   (elan-flatten (mapcar (function (lambda (form) (elan-free-names-entry-of-defining-form form file-path)))
           form-list-ok))))

(defun elan-free-names-of-all-forms-outside-file (file-path)
  "Return the list of free name entries of all definitions in file-path.
Remove those free names that belong to the definitions in this file."
  (let* ((fnl (elan-free-names-of-all-forms file-path))
         (defined-names (mapcar (function elan-defining-name-of-free-name-entry) fnl))
        )
    (mapcar
      (function (lambda (fn-entry) 
                  (elan-make-free-name-entry 
                    (elan-defining-name-of-free-name-entry fn-entry)
                    (elan-kind-of-free-name-entry fn-entry)
                    (elan-file-path-of-free-name-entry fn-entry)
                    (elan-list-difference (elan-free-name-list-of-free-name-entry fn-entry) defined-names)
                  )
                )
      )
      fnl)))

(defun elan-free-names-of-all-forms-outside-code-base (code-base)
  "Return the list of free name entries of all definitions in the code-base file-paths.
Remove those free names that belong to the definitions in the analysis code base. Thus, this function 
returns a list similar to the list returned by free-names-of-all-forms-all-paths. But there are fewer 
free names in the list returned by free-names-of-all-forms-outside-code-base,
because names which are defined within the analysis code are not considered as free names."
 (let* ((fnl (elan-free-names-of-all-forms-all-paths code-base))
        (defined-names (mapcar (function elan-defining-name-of-free-name-entry) fnl))
       )
    (mapcar
      (function (lambda (fn-entry) 
                  (elan-make-free-name-entry 
                    (elan-defining-name-of-free-name-entry fn-entry)
                    (elan-kind-of-free-name-entry fn-entry)
                    (elan-file-path-of-free-name-entry fn-entry)
                    (elan-list-difference (elan-free-name-list-of-free-name-entry fn-entry) defined-names)
                  )
                )
      )
      fnl)))


    
  
(defun elan-free-names-entry-of-defining-form (form file-path)
 "Construct and finds the free names of a top-level, defining form in file-path. Return a free name entry of the form."
 (if (elan-defining-form-p form)
     (if (elan-setq-form-p form)
         (let* ((symbol-expr-list (cdr form))  ; an even length list
                (symbol-list (elan-every-second-element symbol-expr-list))
                (expr-list   (elan-every-second-element (cdr symbol-expr-list))))
           (elan-mapcar2 (function (lambda (sym expr) (elan-make-free-name-entry sym 'setq  file-path (elan-free-names-of-form expr)))) symbol-list expr-list))
        (list (elan-make-free-name-entry (elan-name-of-defining-form form) (elan-kind-of-defining-form form)  file-path (elan-free-names-of-form form))))
     (list (list 'PROBLEM form))))

(defun elan-duplicate-definitions (free-name-entries)
  "Return those free-name-entries that are duplicated with respect to its defining name."
  (let ((duplicated-defining-names (elan-duplicates (mapcar (function elan-defining-name-of-free-name-entry) free-name-entries))))
    (elan-flatten 
      (mapcar
        (function (lambda (ddn) (elan-find-all-elements-by-selector ddn free-name-entries (function elan-defining-name-of-free-name-entry))))
        duplicated-defining-names))))



; -------------------------------------------------
; Constructor and selectors of free name entries:

(defun elan-make-free-name-entry (defining-name kind-of-definition file-path free-names)
  (cons defining-name (cons kind-of-definition (cons file-path free-names))))

(defun elan-defining-name-of-free-name-entry (fne)
  (car fne))

(defun elan-kind-of-free-name-entry (fne)
  (car (cdr fne)))

(defun elan-file-path-of-free-name-entry (fne)
  (car (cdr (cdr fne))))

(defun elan-free-name-list-of-free-name-entry (fne)
  (cdr (cdr (cdr fne))))

; -------------------------------------------------


(defun elan-defining-form-p (form)
  "Is form considered a defining form"
  (and (consp form) (consp (cdr form))
       (memq (car form) elan-list-of-toplevel-defining-forms )))

(defun elan-setq-form-p (form)
  "Is form considered a setq form"
  (and (consp form) (consp (cdr form))
       (eq (car form) 'setq )))

(defun elan-kind-of-defining-form (form)
  "Given a defining form, form, return a symbol that classifies it. 
Such as defun, dervar, setq, ..."
  (car form))

(defun elan-name-of-defining-form (form)
  "Assume that form is a defining form. Return the name of it"
  (elan-cadr form))

(defun elan-free-names-of-form (form)
  "Return the list of free names in the Emacs Lisp form form (the first parameter).
Disregards byte code functions, build in functions (subrs), and simple aliases.
Thus, this function does not really give the true result.
See the function all-free-names-of-form."
  (elan-remove-alias-functions
   (elan-remove-byte-code-functions
    (elan-remove-subrps 
     (elan-all-free-names-of-form form)))))

(defvar elan-name-collection '())

(defvar elan-name-collection-p nil 
  "Are we collecting all names encountered in the variable 
name-collection during the traversal in the function  all-free-names-of-form")

(defun elan-all-free-names-of-form (form)
  "Return the list of free names in the Emacs Lisp (top-level) form (the first parameter).
The true result, but without duplicates."
  (setq elan-name-collection '())
  (let ((res (elan-remove-duplicates (elan-free-names-of-form-1 form nil 'normal))))
    (setq elan-name-collection (reverse elan-name-collection))
    res))

(defun elan-name-occurrences-of-form (form)  
  "Return the list all name occurrences of form. Each name occurrence is an association list that maps
name occurrences (symbols/literal strings) to their status (free/non-free/top-level-defining-name).
This function is an alternative use of the function free-names-of-form-1, namely in a mode where
the boolean variable name-collection-p is true."
  (setq elan-name-collection '())
  (setq elan-name-collection-p t)
  (let ((res (elan-free-names-of-form-1 form nil 'normal))
        (real-res (reverse elan-name-collection))
       )
      (setq elan-name-collection '())
      (setq elan-name-collection-p nil)
    (elan-top-level-modification real-res)))

(defun elan-top-level-modification (name-occurrence-list)
  "If name-occurrence-list represents a top level definition (defun, defvar, ...) modify (by mutation) the appropriate
element (typically the second) to be a top-level-defining-name instead of a non-free name. Return name-occurrence-list."
  (if (>= (length name-occurrence-list) 2)
      (cond 
;            ((eq (car (car name-occurrence-list)) 'setq)
;               (mutate-all-symbols-as-top-level-defining-forms (cdr name-occurrence-list)))

            ((memq (car (car name-occurrence-list)) elan-list-of-toplevel-defining-forms)
              (let ((second-el (elan-cadr name-occurrence-list)))
               (setcdr second-el 'top-level-defining-name)))

            (t 'do-nothing)))
   name-occurrence-list)

; NOT used.
; mutate the symbol position (v1, ..., vn) is an assignment form, such as (setq v1 e1 v2 e2 ... vn en)
; to be top-level-definint-name. symbol-expr-name-occurrence-list is the list corresponding to (v1 e1 v2 e2 ... vn en).
(defun elan-mutate-all-symbols-as-top-level-defining-forms (symbol-expr-name-occurrence-list)
 (let ((symbol-expr-name-occurrence-list-1 symbol-expr-name-occurrence-list))
  (while (not (null symbol-expr-name-occurrence-list-1))
    (setcdr (car symbol-expr-name-occurrence-list-1) 'top-level-defining-name)
    (setq symbol-expr-name-occurrence-list-1 (elan-cddr symbol-expr-name-occurrence-list-1)))))

(defun elan-free-names-of-form-1 (form bound-names mode)
  "The function doing the real work of all-free-names-of-form. Recurses through all constituents of form
and collects free names. The parameter bound-names contains those names in form which are binding name occurrences.
The parameter mode has two possible values: normal and constant. If constant, all encountered names are non-free.
If normal, the names are anlalyzed against bound-names. The mode parameter is need by the handling of backquote.
This function has, in fact, two purposes: First, it returns the free names of form relative to bound-names. Second,
it defines the variable name-collection via the functions collect-non-free-name and collect-free-name. The variable name-collection
is used for systematic renaming purposes via the function name-occurrences-of-form."
  (cond ((and (elan-defun-form-p form) (eq mode 'normal)) 
           (let ((defined-name (elan-name-of-defining-form form))
                 (formal-parameter-names (elan-defun-formal-parameters form))
                 (filtered-formal-parameter-names (elan-defun-formal-parameters-filtered form))  ;  without &optional and &rest
                )
              (elan-collect-non-free-name 'defun) 
              (elan-collect-non-free-name defined-name) 
              (mapcar (function (lambda (fp) (elan-collect-non-free-name fp))) formal-parameter-names)
 
              (elan-free-names-of-form-1 
                 (elan-body-of-defun form)
                 (append filtered-formal-parameter-names bound-names)  ; implies that recursive calls count as free names in form
                 mode
              )))

        ((and (elan-defsubst-form-p form) (eq mode 'normal)) 
           (let ((defined-name (elan-name-of-defining-form form))
                 (formal-parameter-names (elan-defun-formal-parameters form))
                 (filtered-formal-parameter-names (elan-defun-formal-parameters-filtered form))  ;  without &optional and &rest
                )
              (elan-collect-non-free-name 'defsubst) 
              (elan-collect-non-free-name defined-name) 
              (mapcar (function (lambda (fp) (elan-collect-non-free-name fp))) formal-parameter-names)
 
              (elan-free-names-of-form-1 
                 (elan-body-of-defun form)
                 (append filtered-formal-parameter-names bound-names)  
                 mode
              )))

         ((and (elan-defmacro-form-p form) (eq mode 'normal))                      ; Macros are not really analyzed. The macro body is treated as a constant, much like a quote expression  
           (let ((defined-name (elan-name-of-defining-form form))
                 (formal-parameter-names (elan-defun-formal-parameters form))
                 (filtered-formal-parameter-names (elan-defun-formal-parameters-filtered form))  ;  without &optional and &rest
                )
              (elan-collect-non-free-name 'defmacro) 
              (elan-collect-non-free-name defined-name) 
              (mapcar (function (lambda (fp) (elan-collect-non-free-name fp))) formal-parameter-names)
 
              (elan-free-names-of-form-1 
                 (elan-body-of-defmacro form)
                 (append filtered-formal-parameter-names bound-names)  ; never mind - does not really matter  
                 'constant
              )))

        ((and (elan-defvar-form-p form) (eq mode 'normal)) 
           (let ((defined-name (elan-name-of-defining-form form))
                )
              (elan-collect-non-free-name 'defvar) 
              (elan-collect-non-free-name defined-name) 
              (elan-free-names-of-form-1 (elan-body-of-defvar form) bound-names mode)))

        ((and (elan-defconst-form-p form) (eq mode 'normal)) 
           (let ((defined-name (elan-name-of-defining-form form))
                )
              (elan-collect-non-free-name 'defconst) 
              (elan-collect-non-free-name defined-name) 
              (elan-free-names-of-form-1 (elan-body-of-defconst form) bound-names mode)))

        ((and (elan-defcustom-form-p form) (eq mode 'normal)) 
           (let ((defined-name (elan-name-of-defining-form form))
                )
              (elan-collect-non-free-name 'defcustom) 
              (elan-collect-non-free-name defined-name) 
              (elan-free-names-of-form-1 (elan-body-of-defcustom form) bound-names mode)))

        ((and (elan-lambda-form-p form) (eq mode 'normal))
            (let ((lambda-parameters (elan-lambda-formal-parameters form))
                  (filtered-lambda-parameters (elan-lambda-formal-parameters-filtered form)))
              (elan-collect-non-free-name 'lambda)
              (mapcar (function (lambda (fp) (elan-collect-non-free-name fp))) lambda-parameters)
              (elan-free-names-of-form-1 (elan-cddr form) (append filtered-lambda-parameters bound-names) mode)))
 
        ((and (elan-let-form-p form) (eq mode 'normal))
           (let* ((binding-part-of-let (elan-cadr form))
                  (let-names (mapcar (function (lambda (binding) (elan-name-of-let-binding binding))) binding-part-of-let))
                  (let-exprs (mapcar (function (lambda (binding) (elan-expressions-of-let-binding binding))) binding-part-of-let))
                  (let-body (elan-cddr form))
                 )
              (elan-collect-non-free-name 'let)
              (append
                (elan-flatten 
                  (elan-mapcar2
                    (function (lambda (name expr) 
                                (elan-collect-non-free-name name)
                                (if (null expr) ; only name, no initial expr
                                    nil
                                    (elan-free-names-of-form-1 expr bound-names mode))))
                    let-names let-exprs))
                (elan-free-names-of-form-1 (elan-cddr form) (append let-names bound-names) mode))))

        ((and (elan-let-star-form-p form) (eq mode 'normal))
           (let* ((binding-part-of-let (elan-cadr form))
                  (let-names (mapcar (function (lambda (binding) (elan-name-of-let-binding binding))) binding-part-of-let))
                  (let-exprs (mapcar (function (lambda (binding) (elan-expressions-of-let-binding binding))) binding-part-of-let))
                  (let-body (elan-cddr form))
                 )
              (elan-collect-non-free-name 'let*)
              (append
                (elan-bounded-let-star-names let-names let-exprs bound-names)
                (elan-free-names-of-form-1 (elan-cddr form) (append let-names bound-names) mode))))

        ((and (elan-quote-form-p form) (eq mode 'normal))
            (elan-collect-non-free-name 'quote)
            (elan-free-names-of-form-1 (cdr form) bound-names 'constant)
        )


        ((and (elan-backquote-form-p form)  (eq mode 'normal)) 
            (elan-collect-non-free-name elan-el-backquote-symbol)
            (elan-free-names-of-form-1 (cdr form) bound-names 'semi-constant))

        ((and (elan-unquote-form-p form) (eq mode 'semi-constant)) 
            (elan-collect-non-free-name elan-el-unquote-symbol)
            (elan-free-names-of-form-1 (cdr form) bound-names 'normal))

        ((and (elan-unquote-splice-form-p form) (eq mode 'semi-constant)) 
            (elan-collect-non-free-name elan-el-unquote-splice-symbol)
            (elan-free-names-of-form-1 (cdr form) bound-names 'normal))


        ((and (elan-function-form-p form)  (eq mode 'normal)) 
            (elan-collect-non-free-name 'function)
            (elan-free-names-of-form-1 (elan-cadr form) bound-names mode))

        ((symbolp form) (cond ((keywordp form) 
                                (progn
                                   (elan-collect-non-free-name form)
                                   nil))
                              ((and (eq mode 'normal) (not (member form bound-names)) (not (member form (list 'nil 't))))
                                (progn
                                  (elan-collect-free-name form)
                                  (list form)))
                              ((member form (list 'nil 't))
                                (progn
                                  (elan-collect-free-name form)
                                  nil))
                              (t (progn
                                   (if form (elan-collect-non-free-name form))
                                   nil
                                   ))))

;        ((numberp form) (collect-non-free-name form)
;                         nil)

        ((consp form) (append (elan-free-names-of-form-1 (car form) bound-names mode)
                              (if (not (null (cdr form)))
                                  (elan-free-names-of-form-1 (cdr form) bound-names mode)
                                nil)))

        ((vectorp form)   
           (elan-free-names-of-form-1 (elan-vector-to-list form) bound-names 'constant))

        ((stringp form)
           (elan-collect-non-free-name form)
           nil)
 
        (t nil)))



(defun elan-collect-free-name (name)
  (if elan-name-collection-p
    (setq elan-name-collection (cons (cons name 'free)  elan-name-collection))))

(defun elan-collect-non-free-name (name)
  (if elan-name-collection-p
    (setq elan-name-collection (cons (cons name 'non-free)  elan-name-collection))))

(defun elan-name-of-let-binding (binding)
  (cond ((symbolp binding) binding)
        ((consp binding) (car binding))
        (t (error "Illegal format of let binding"))))

(defun elan-expressions-of-let-binding (binding)
  (cond ((symbolp binding) nil)
        ((consp binding) (cdr binding))
        (t (error "Illegal format of let binding"))))

(defun elan-body-of-defun (defun-form)
  (cdr (cdr (cdr defun-form))))

(defun elan-body-of-defmacro (defmacro-form)
  (cdr (cdr (cdr defmacro-form))))

(defun elan-body-of-defvar (defvar-form)
  (cdr (cdr defvar-form)))

(defun elan-body-of-defconst (defconst-form)
  (cdr (cdr defconst-form)))

(defun elan-body-of-defcustom (defcustom-form)
  (cdr (cdr defcustom-form)))

(defun elan-remove-subrps (names)
  (elan-filter (function (lambda (n) (not (elan-is-subr n)))) names))

(defun elan-remove-byte-code-functions (names)
  (elan-filter (function (lambda (n) (not (elan-is-byte-code-function n)))) names))

(defun elan-remove-alias-functions (names)
  (elan-filter (function (lambda (n) (not (elan-is-alias-function n)))) names))

(defun elan-is-subr (name)
  (if (fboundp name)
      (subrp (symbol-function name))
      nil))

(defun elan-is-byte-code-function (name)
  (if (fboundp name)
      (byte-code-function-p (symbol-function name))
      nil))

(defun elan-is-macro (name)
  (if (fboundp name)
      (let ((fval (symbol-function name)))
         (and (consp fval) (eq (car fval) 'macro)))
      nil))

(defun elan-is-alias-function (name)
  (if (fboundp name)
      (symbolp (symbol-function name))
      nil))  



(defun elan-bounded-let-star-names (let-names let-exprs bound-names)
  (if (null let-names)
      nil
      (let ((let-name (car let-names))
            (let-expr (car let-exprs)))
        (elan-collect-non-free-name let-name)
        (if (null let-expr)
            (elan-bounded-let-star-names (cdr let-names) (cdr let-exprs) (cons let-name bound-names))
            (append (elan-free-names-of-form-1 let-expr bound-names 'normal)
                    (elan-bounded-let-star-names (cdr let-names) (cdr let-exprs) (cons let-name bound-names)))))))



(defun elan-defun-formal-parameters (defun-form)
  (elan-caddr defun-form))

(defun elan-defun-formal-parameters-filtered (defun-form)
  (elan-filter (function (lambda (p) (not (memq p '(&optional &rest))))) (elan-caddr defun-form)))

(defun elan-lambda-formal-parameters (lambda-form)
   (elan-cadr lambda-form))

(defun elan-lambda-formal-parameters-filtered (lambda-form)
  (elan-filter (function (lambda (p) (not (memq p '(&optional &rest))))) (elan-cadr lambda-form)))

(defun elan-defvar-form-p (form)
  (and (consp form) (eq (car form) 'defvar)))

(defun elan-defconst-form-p (form)
  (and (consp form) (eq (car form) 'defconst)))

(defun elan-defcustom-form-p (form)
  (and (consp form) (eq (car form) 'defcustom)))

(defun elan-defun-form-p (form)
  (and (consp form) (eq (car form) 'defun)))

(defun elan-defsubst-form-p (form)
  (and (consp form) (eq (car form) 'defsubst)))

(defun elan-defmacro-form-p (form)
  (and (consp form) (eq (car form) 'defmacro)))

(defun elan-let-form-p (form)
  (and (consp form) (eq (car form) 'let)))

(defun elan-let-star-form-p (form)
  (and (consp form) (eq (car form) 'let*)))

(defun elan-quote-form-p (form)
  (and (consp form) (eq (car form) 'quote)))

; To avoid gnat in this el source program:
(defvar elan-el-backquote-symbol '\`)
(defvar elan-el-unquote-symbol '\,)
(defvar elan-el-unquote-splice-symbol '\,@)

(defun elan-backquote-form-p (form) 
  (and (consp form) (eq (car form) elan-el-backquote-symbol)))

(defun elan-unquote-form-p (form)
  (and (consp form) (eq (car form) elan-el-unquote-symbol)))

(defun elan-unquote-splice-form-p (form)
  (and (consp form) (eq (car form) elan-el-unquote-splice-symbol)))

(defun elan-function-form-p (form)
  (and (consp form) (eq (car form) 'function)))

(defun elan-lambda-form-p (form)
  (and (consp form) (eq (car form) 'lambda)))



 
(defun elan-file-read-code-base (code-base-file-path)
  "Return a linear list of (absolute) file paths from code-base-file-path.
The parameter code-base-file-path is the file path to a cbs file, which enumerates el source code files.
code-base-file-path may, in a recursive way, contain code base files (-.cbs files) which are expanded by this function."
 (let ((code-base-initial-path (elan-ensure-trailing-slash (file-name-directory code-base-file-path))))
  (let* ((code-base-list-0 (elan-file-read code-base-file-path))
         (code-base-list (mapcar (function (lambda (fp) (elan-make-absolute-file-path fp code-base-initial-path))) code-base-list-0))
       )
    (elan-accumulate-right
      (function append)
      nil
      (mapcar 
       (function
        (lambda
          (code-base-element)
          (cond ((member (elan-file-name-extension code-base-element) (list "el" "EL")) (list code-base-element))   ; notice list surrounding
                ((equal (elan-file-name-extension code-base-element) "cbs") (elan-file-read-code-base code-base-element))
                (t (error "Code base files (cbs files) are only allowed to contain el or cbs files."))))) 
       code-base-list)))))

(defun elan-make-absolute-file-path (file-path prefix)
 (elan-eliminate-tilde-prefix
  (if (elan-absolute-file-path-p file-path)
      file-path
      (concat prefix file-path))))

(defun elan-find-el-definition (name kind file-path)
  "Find and return the definition (a string) of name of kind in file-path"
  (let* ((buf-name "program")
         (program-buffer (if (get-buffer buf-name) (get-buffer buf-name) (generate-new-buffer buf-name)))
         (res nil)
         (def-exp (list 'at-beginning-of-line
                            (list 'concat "(" '(zero-or-more (char-set (char 32))) (elan-symbol-to-string kind)
                                          '(zero-or-more (char-set (char 32))) (elan-symbol-to-string name) 
                                          '(one-or-more (char-set (char 32) (char 10) (char 13) (char 40))))))
         (def-reg-exp (elan-regular-expression def-exp))
        )
    (save-excursion
      (set-buffer program-buffer)
      (erase-buffer)
      (insert-file file-path)
      (goto-char (point-min))
      (let ((search-res (re-search-forward def-reg-exp nil t)))  ; too simple - use regexp
        (if search-res
            (let ((p nil))
              (beginning-of-defun 1)
              (setq p (point))
              (forward-sexp 1)
              (setq res (buffer-substring-no-properties p (point))))
            (setq res nil)))
      (kill-buffer program-buffer))
    res)) 
            

(defun elan-insert-free-name-report (free-name-lst)
  (insert 
   (elan-list-to-string
     (mapcar
      (function 
       (lambda (fn) 
          (concat 
            (elan-symbol-to-string fn) 
            (if (boundp fn) "[Variable]" "")) ))
      free-name-lst)
    " ")))


; ---------------------------------------------------------------------------------------------------------------
; Name replacing stuff.

(defvar elan-name-replacement-list nil "A list of name replacements which can be carried out")

(defun elan-do-rename-all-forms-in-current-buffer ()
  (goto-char (point-min))
  (elan-forward-through-white-space-and-comments)
  (while (not (eobp))
    (progn 
      (elan-do-rename-in-single-form-under-point)
      (elan-forward-through-white-space-and-comments))))

(defun elan-do-rename-in-single-form-under-point ()
  "Substitute free names in the form under point in the current buffer, according to the global name-replacement-list
Following the renamings, leave point after the form."
   (let* ((parsed-form (elan-read-current-form (current-buffer)))
          (name-occurrences (elan-name-occurrences-of-form parsed-form))
          (end-point (elan-end-point-of-current-form))  ; not useful, because replacements actually alter this end point...
         )
     (while (not (null name-occurrences))
       (let* ((next-name-occurrence (car name-occurrences))
              (next-name (car next-name-occurrence))
              (next-kind (cdr next-name-occurrence))
              (name-replacement (assq next-name elan-name-replacement-list))
              (message-id (elan-message-identification-of-form parsed-form))
             )
         (elan-message-about-form message-id)
         (elan-locate-next-name next-name next-kind end-point message-id)
         (if (and (or (eq next-kind 'free) (eq next-kind 'top-level-defining-name))   ; free name, or top level defining name,
                  (symbolp next-name)                                                 ; ... not a string,
                  name-replacement                                                    ; ... for which there is a name replacement,
                  (not (eq next-name (elan-cadr name-replacement)))                        ; ... which is NOT the identify replacement
             )
             (progn (backward-delete-char (length (symbol-name next-name))) (insert (elan-symbol-to-string (elan-cadr name-replacement)))))

         (setq name-occurrences (cdr name-occurrences))))
      (beginning-of-defun 1)
      (forward-sexp 1)))

(defun elan-read-current-form (buffer)
  (save-excursion 
    (read buffer)))

(defun elan-end-point-of-current-form ()
  (save-excursion
    (forward-sexp 1)
    (point)))

(defconst elan-empty-parenthesis-pair-reg-exp '(concat "(" (zero-or-more (char-set (char 32) (char 10) (char 13) (char 40))) ")") "Regular expression for empty parenthesis pair")

(defun elan-locate-next-name (name kind limit-point-of-search error-message-name)
  "Match name by searching for it. name may also be a literal string from the program.
It is a fatal error if name is not found as a complete word (not just a prefix).
Leave point after the matched name. The parameter limit-point-of-search is not used (because name substitutions alter the ending point).
error-message-name is a name of the surrounding form, used solely for error message purposes."
  (cond ((stringp name)
           (let ((search-res (elan-search-forward-not-comment "\"")))  
             (if search-res
                 (progn (backward-char 1) (forward-sexp 1))  ; just skip string - do not attempt more specific matching
                 (error (concat "Cannot find string: " name )))))

        ((and (symbolp name) (or (eq name 'quote) (eq name elan-el-backquote-symbol) (eq name elan-el-unquote-symbol) (eq name elan-el-unquote-splice-symbol) ))
             ; ignore - not necessarily safe to do so. Next call of locate-next-name most likely remedies the lacking matching 
             'do-nothing
        )

        ((and (symbolp name) (eq name nil))
           (let* ((search-res (elan-multi-search-forward-not-comment (list "nil" elan-empty-parenthesis-pair-reg-exp))))
             (forward-sexp -1)
             (let ((delim (elan-sexp-delimitor)))
               (if (or (not search-res) ; name not found
                       (not (= (- search-res 1) (elan-cadr delim)))) ; not match of some prefix of name
                   (error (concat "Mismatch: " "The empty list or nil not found in " error-message-name))
                   (goto-char search-res)))))

        ((symbolp name)
           (let* ((search-res (elan-search-forward-not-comment (elan-symbol-to-string name))))
             (forward-sexp -1)
             (let ((delim (elan-sexp-delimitor)))
               (if (or (not search-res) ; name not found
                       (not (= (- search-res 1) (elan-cadr delim)))) ; not match of some prefix of name
                   (error (concat "Mismatch: " (elan-symbol-to-string name) " not found in " error-message-name))
                   (goto-char search-res)))))

        ((numberp name)   ; This case is not used.  Problem and complication:  Numbers also cover character notation, such as ?a.
           (let* ((search-res (elan-search-forward-not-comment (number-to-string name)))
                 )
             (forward-sexp -1)
             (let ((delim (elan-sexp-delimitor)))
               (if (or (not search-res) ; name not found
                       (not (= (- search-res 1) (elan-cadr delim)))) 
                   (error (concat "Mismatch: " (number-to-string name) " not found in " error-message-name))
                 (goto-char search-res)))))

        (t (error (concat "locate-next-name: Should not happen. " error-message-name)))))

(defun elan-search-forward-not-comment (str)
  (let ((search-res (search-forward str nil t)))
    (if (not (elan-is-inside-comment))
        search-res
        (elan-search-forward-not-comment str))))

(defun elan-multi-search-forward-not-comment (search-list)
  (let ((search-res (elan-multi-search-forward search-list)))
    (if (not (elan-is-inside-comment))
        search-res
        (elan-multi-search-forward search-list))))

(defun elan-message-identification-of-form (form)
    (let ((name (cond ((elan-defun-form-p form) 
                       (elan-symbol-to-string (elan-name-of-defining-form form)))
                      ((elan-defsubst-form-p form) 
                       (elan-symbol-to-string (elan-name-of-defining-form form)))
                      ((elan-defvar-form-p form) 
                        (elan-symbol-to-string (elan-name-of-defining-form form)))
                      ((elan-defconst-form-p form) 
                         (elan-symbol-to-string (elan-name-of-defining-form form)))
                      ((elan-defcustom-form-p form) 
                         (elan-symbol-to-string (elan-name-of-defining-form form)))
                      (t "another form"))))
       name))

(defun elan-message-about-form (about-name)
  "Issue a short message (in minibuffer, via the message function) about form, which is identified of about-name. For progress information during renaming"
    (message (concat "Renaming: " about-name)))

    
; ===============================================================================================================

(defun elan-current-directory ()
  "Return the directory of the current buffer, if such a directory makes sense
for the buffer"
  (if (eq major-mode 'dired-mode)
      (dired-current-directory)
      (elan-buffer-directory-name (current-buffer))))

(defun elan-string-to-symbol (str)
  "Convert STR to a symbol."
 (car (read-from-string str)))

(defvar elan-CR "
" "CR character constant")

(defun elan-file-read-all (file)
  "Read the list of all Lisp expressions from file (full path)"
  (let ((res nil))
   (save-excursion
     (let ((temp-buf (generate-new-buffer "file-reading.tmp")))
       (set-buffer temp-buf)
       (insert-file file)
       (goto-char (point-min))
       (elan-forward-through-white-space-and-comments)

       (while (not (eobp))
         (progn 
          (setq res (cons (read temp-buf) res))
          (elan-forward-through-white-space-and-comments)))
       (kill-buffer temp-buf)
       (reverse res)))))

(defun elan-list-difference (l1 l2)
  "Return the list difference between l1 and l2. Thus, return all those elements in
l1 which are not member of l2. For an elemeng i l1, if it belongs to l2, remove it from l1.
Else include it. A non-destructive function. Comparison done by equal via member."
  (elan-filter 
    (function (lambda (e1) (not (member e1 l2))))
    l1))

(defun elan-every-second-element (lst)
  "Return the first, third, etc element of the list lst until it is exhausted."
  (let ((lst1 lst) (res nil) (i 0))
    (while lst1
       (if (evenp i) (setq res (cons (car lst1) res)))
       (setq i (+ i 1))
       (setq lst1 (cdr lst1)))
    (reverse res)))

(defun elan-duplicates (lst)
  "Return the elements in lst which occur two or more times. Non-destructive.
Comparison done by equal."
  (let ((the-lst lst)
        (res nil)
       )
    (while (not (null the-lst))
           (progn
              (if (member (car the-lst) (cdr the-lst))
                  (setq res (cons (car the-lst) res)))
              (setq the-lst (cdr the-lst))))
    (elan-remove-duplicates (reverse res)) ; if remove-duplicates is not called, some duplicates may be reported more than once
  ))

(defun elan-find-all-elements-by-selector (e lst selector)
  "Find all elements x in list for which (equal (selector x) e)." 
  (let ((the-lst lst)
        (res nil))
    (while (not (null the-lst))
        (progn
          (if (equal (apply selector (list (car the-lst) )) e)
              (setq res (cons (car the-lst) res)))
          (setq the-lst (cdr the-lst))))

   (reverse res)))

(defun elan-remove-duplicates (lst)
  "Remove the duplicate from lst and return the result. Non-destructive.
Comparison done by equal."
  (let ((the-lst lst)
        (res nil)
       )
    (while (not (null the-lst))
           (progn
              (if (not (member (car the-lst) (cdr the-lst)))
                  (setq res (cons (car the-lst) res)))
              (setq the-lst (cdr the-lst))))
    (reverse res)))

(defun elan-flatten (list-of-lists)
  (elan-accumulate-right (function append) nil list-of-lists))

(defun elan-mapcar2 (f x y)
 "Generalization of mapcar.
Do the mapping as long as there are elements in both lists. Programmed iteratively."
  (let ((xlst x) (ylst y) 
        (res nil))
    (while (not (or (null xlst) (null ylst)))
      (setq res (cons (apply f (list (car xlst) (car ylst))) res))
      (setq xlst (cdr xlst)) (setq ylst (cdr ylst)))
    (reverse res)))

(defun elan-cddr (l)
  (cdr (cdr l)))

(defun elan-vector-to-list (vector)
  "Return the list of elements of vector"
  (let ((lgt (length vector))
        (i 0)
        (res nil))
    (while (< i lgt)
      (setq res (cons (elt vector i) res))
      (setq i (+ i 1)))
    (reverse res)))

(defun elan-caddr(l)
  (car (cdr (cdr l))))

(defun elan-filter (p x)
"Filter takes two arguments: a predicate P and a list X.
Return the elements of the list X that satiesfy the predicate P"
;;Non-recursive version of filter. Overwrites the previous recursive one.
  (let ((res-list nil))
    (while x
      (if (apply p (list (car x)))
          (setq res-list (cons (car x) res-list)))
      (setq x (cdr x)))
    (reverse res-list)))

(defun elan-ensure-trailing-slash (dir-string)
  "Ensure that dir-string ends in a slash: /"
  (let* ((dir-lgt (length dir-string))
         (last-chr (aref dir-string (- dir-lgt 1))))
   (if (eq last-chr ?/)
       dir-string
       (concat dir-string "/"))))

(defun elan-file-read (file)
  "Read the first Lisp expression from file (full path)"
  (save-excursion
    (let ((temp-buf (generate-new-buffer "file-reading.tmp")))
      (set-buffer temp-buf)
      (insert-file file)
      (goto-char (point-min))
      (prog1
        (read temp-buf)
        (kill-buffer temp-buf)))))

(defun elan-accumulate-right (f startval lst)
  "Reduce the list lst by the binary function f.
   Use start value startval. Programmed iteratively"
  (let ((lst-rev (reverse lst))
        (res startval))
    (while lst-rev
      (setq res (apply f (list (car lst-rev) res)))
      (setq lst-rev (cdr lst-rev)))
    res))

(defun elan-file-name-extension (filename)
  "Takes a filename, without directory path.
Returns the extension. If no extension is found, return nil."
  (let ((point-idx (elan-find-last-point filename)))
    (if point-idx 
        (substring filename (+ 1 point-idx) (length filename))
        nil)))

(defun elan-eliminate-tilde-prefix (dir-path)
 "If dir-path is initiated with a tilde char ~/ prefix substitute it with the home directory of the current user.
This is only supposed to happen on Unix systems, in which case we rely on the environment HOME variable to get information about
the home dir string."
 (if (= ?~ (aref dir-path 0))
     (concat (elan-user-home-dir) (substring dir-path 2 (length dir-path)))
     dir-path))

(defun elan-absolute-file-path-p (p)
  "Is the file path p an absolute file path"
  (or (elan-absolute-windows-file-path-p p) (elan-absolute-unix-file-path-p p)))

(defun elan-regular-expression (reg-exp)
  "Return a textual regular expression from the structural
regular-expression REG-EXP. This function is the external interface
function of the structural regular-expression package. The following
summarizes the syntax. X-es are structural regular expressions,
including strings. C-es are characters which we represent as string of
length one. Within string X-es the characters that have
special meanings within textual regular expressions can be used
without problems (they are explicitly translated). Similarly
with C-es in character sets the character strings `]', `-' and `^',
which are special in textual regular expressions, can be used
without problems, because proper translations are performed.

Syntax symmary:

  (group X1)  
  (concat X1 ... Xn)
  (or X1 ... Xn)                                     
  (any-single-char)                                  
  (zero-or-more X)                                   
  (one-or-more X)                                    
  (zero-or-one X)                                    

  (char-set C1 ... CN)                               
  (complement-char-set C1 ... CN)                    
  (char-range C1 CN)                                 
  (complement-char-range C1 CN)                      
  Ci may also be:
    tab       backspace
    newline   vertical-tab
    formfeed  return
    escape    space
    (char n), n a decimal number

  (at-beginning-of-line X)                           
  (at-end-of-line X)                                 
  (at-beginning-of-buffer X)                         
  (at-end-of-buffer X)                               
  (at-beginning-of-word X)                           
  (at-end-of-word X)                                 
  (any-single-word-constituent-char)                 
  (any-single-non-word-constituent-char)             

  (syntax-code i),  
  (not-syntax-code i),
    i is a character syntax code,
    which either is a one letter string
    or one of the symbols
      whitespace            word-constituent
      symbol-constituent    punctuation-character
      open-parenthesis      close-parenthesis
      close-parenthesis     string-quote
      escape                character-quote
      paired-delimiter      expression-prefix
      comment-starter       comment-ender
"
  
  (elan-reg-exp-traverse reg-exp))

(defun elan-list-to-string (lst &optional sep)
 "Convert a list of mixed chars and strings to a string. Separate elements by the string sep."
  (let* ((sep1 (if sep sep ""))
         (lgt-lst (length lst))
         (lgt-sep1 (length sep1)))
    (let ((str (make-string (+ (* (max (- lgt-lst 1) 0) lgt-sep1) (elan-length-char-string-list lst)) 0))
          (lst1 lst)
          (i 0) (j 0))
      (while lst1 
        (elan-aset-char-string str i (car lst1))
        (setq i (+ i (elan-length-char-string (car lst1)))) ; advance i by element length
        (if (< j (- lgt-lst 1)) 
            (progn (elan-aset-char-string str i sep1) ; insert separator
                   (setq i (+ i lgt-sep1))))     ; advance i by separator length
        (setq j (+ j 1)) 
        (setq lst1 (cdr lst1)))
      str)))

(defun elan-forward-through-white-space-and-comments ()
 "Move point over white space and comments in an elisp file"
 (while (or (elan-looking-at-white-space (point)) 
            (elan-looking-at-chars (point) (list 59)))   ; looking at semicolon
     (cond ((elan-looking-at-white-space (point)) (forward-char 1))
           ((elan-looking-at-chars (point) (list 59)) (forward-line 1)))))

(defun elan-sexp-delimitor ()
  "Return a list of two numbers: the start and the end of the word under point"
 (save-excursion
  (let ((re (progn (forward-sexp 1)
                 (point)))
        (rb (progn (forward-sexp -1)
                 (point))))
    (if (and rb re) (list rb (- re 1))))))

(defun elan-cadr (l)
  (car (cdr l)))

(defun elan-is-inside-comment ()
  "Is point inside a string"
  (save-excursion
    (let ((pt (point)))
      (beginning-of-defun 1)
      (let* ((parse-result (parse-partial-sexp (point) pt))
             (inside-comment-info (nth 4 parse-result)))
        (if inside-comment-info t nil)))))

(defun elan-multi-search-forward (search-list)
  "Search for each element in search-list, starting in the current possition of point.
An element in search-list can be a string or a structural regular expression (a list).
The effect and result of the closest match takes effect.
If no match at all occurs, nil is retured, and point is not affected. 
Else, point is moved to the end of the match, and the value of point after the match is returned."
 (let ((case-fold-search nil))   ; case sensitive search
  (let* ((start (point))
	 (search-result 
	  (mapcar 
	   (function
	    (lambda (search-element)
	      (goto-char start)
	      (cond ((stringp search-element)
		     (search-forward search-element nil t))
		    ((consp search-element)
		     (re-search-forward (elan-regular-expression search-element) nil t))
		    (t (error "multi-search: Unknown kind of element in search list"))))) 
	   search-list))
	 (number-search-result (elan-filter (function numberp) search-result))
	 (min-search-result (if (null number-search-result) nil (apply (function min) number-search-result))))
    (if min-search-result 
        (progn
          (goto-char min-search-result)
          min-search-result)
      nil))))

(defun elan-symbol-to-string (symbol)
  "Convert SYMBOL to a string."
  (symbol-name symbol))

(defun elan-buffer-directory-name (buffer)
  "Return the name of the directory of the file, associated with the
current buffer. If none, return nil"
  (let ((file-name (buffer-file-name buffer)))
    (if file-name
        (file-name-directory file-name)
        nil)))

(defun elan-find-last-point (str)
  "Return the position of the last point in the string str. 
Do not allow to pass '/' while looking for the point. If no point is found under these conditions, return nil."
  (let ((idx (- (length str) 1)))
    (while (and (>= idx 0) (not (eq (aref str idx) ?.)) (not (eq (aref str idx) ?/))) (setq idx (- idx 1)))
    (if (and (>= idx 0) (eq (aref str idx) ?.)) idx nil)))

(defun elan-user-home-dir ()
  "Return a forward slash terminated home directory."
  (let ((home (getenv "HOME")))
    (if (= ?/ (aref home (- (length home) 1)))   ; is home forward slash terminated
        home
        (concat home "/"))))

(defun elan-absolute-windows-file-path-p (p)
  "Is the file path p an absolute file path in windows. Ala c:/xxx/"
  (and (>= (length p) 3) (= ?: (aref p 1)) (= ?/ (aref p 2))))

(defun elan-absolute-unix-file-path-p (p)
  "Is the file path p an absolute file path"
  (= ?/ (aref p 0)))

(defun elan-reg-exp-traverse (reg-exp)
  "Return a textual regular expression from the structural
regular expression REG-EXP-LIST.  This is the toplevel function
in the structural regular expression package."
  (cond ((listp reg-exp) (elan-reg-exp-traverse-list reg-exp))
        ((stringp reg-exp) (elan-reg-exp-traverse-string reg-exp))
        ((symbolp reg-exp) 
         (elan-reg-exp-char-set-list-element reg-exp))
        (t (error "Error in structural regular expression"))))

(defun elan-length-char-string-list (lst)
  "Calculate the length of lst. Chars count 1, strings count their lengths"
 (let ((res 0)
       (lst1 lst))
  (while lst1
    (setq res (+ res (elan-length-char-string (car lst1))))
    (setq lst1 (cdr lst1)))
  res))

(defun elan-length-char-string (x)
  "Return the length of x. Chars count 1, strings count their lengths"
  (cond ((integerp x) 1)
        ((stringp x) (length x))))

(defun elan-aset-char-string (str i x)
  "like aset, but x may also be a string, in which case all chars in x are inserted"
  (cond ((integerp x) (aset str i x))
        ((stringp x) (elan-aset-string str i x 0))))

(defun elan-looking-at-white-space (p)
 (let ((ch (char-after p)))
  (member ch (list 9 10 13 32))))

(defun elan-looking-at-chars (p char-list)
  "is char at possition p member of char-list"
 (let ((ch (char-after p)))
  (member ch char-list)))

(defun elan-reg-exp-traverse-list (reg-exp-list)
  "Handle a structural regular expression list"
  (cond  ((eq 'group  (car reg-exp-list))
              (concat "\\("
                      (elan-reg-exp-traverse (elan-cadr reg-exp-list))
                      "\\)"))
         ((eq 'concat (car reg-exp-list)) 
            (mapconcat
              (function elan-reg-exp-traverse)
              (cdr reg-exp-list)
              ""))
        ((eq 'zero-or-more  (car reg-exp-list))
              (concat "\\("
                      (elan-reg-exp-traverse (elan-cadr reg-exp-list))
                      "\\)"
                      "*"))
        ((eq 'one-or-more  (car reg-exp-list))
              (concat "\\("
                      (elan-reg-exp-traverse (elan-cadr reg-exp-list))
                      "\\)"
                      "+"))
        ((eq 'zero-or-one  (car reg-exp-list))
              (concat "\\("
                      (elan-reg-exp-traverse (elan-cadr reg-exp-list))
                      "\\)"
                      "?"))
        ((eq 'or (car reg-exp-list))
              (mapconcat 
                 (function 
                   (lambda (re)
                      (elan-reg-exp-parenthesize (elan-reg-exp-traverse re))))
                 (cdr reg-exp-list)
                 "\\|"))
       ((eq 'any-single-char (car reg-exp-list)) ".")
       ((eq 'at-beginning-of-line (car reg-exp-list))
              (concat "^\\("
                      (elan-reg-exp-traverse (elan-cadr reg-exp-list))
                      "\\)"))
       ((eq 'at-end-of-line (car reg-exp-list))
              (concat "\\("
                      (elan-reg-exp-traverse (elan-cadr reg-exp-list))
                      "\\)$"))
        ((eq 'char-set  (car reg-exp-list))
              (concat "["
                 (mapconcat
                   (function elan-reg-exp-char-set-element)
                   (elan-reg-exp-global-char-set-translate (cdr reg-exp-list))
                   "")
                 "]"))
        ((eq 'complement-char-set  (car reg-exp-list))
              (concat "[^"
                 (mapconcat
                   (function elan-reg-exp-char-set-element)
                   (elan-reg-exp-global-char-set-translate (cdr reg-exp-list))
                   "")
                 "]"))
        ((eq 'char-range  (car reg-exp-list))
              (concat "["
                       (concat (elan-reg-exp-char-set-element (elan-cadr reg-exp-list))
                                  "-"
                               (elan-reg-exp-char-set-element (elan-caddr reg-exp-list)))
                      "]"))
        ((eq 'complement-char-range  (car reg-exp-list))
              (concat "[^"
                       (concat (elan-reg-exp-char-set-element (elan-cadr reg-exp-list))
                                  "-"
                               (elan-reg-exp-char-set-element (elan-caddr reg-exp-list)))
                      "]"))
        ((eq 'at-beginning-of-buffer (car reg-exp-list))
               (concat "\\`\\("
                      (elan-reg-exp-traverse (elan-cadr reg-exp-list))
                      "\\)"))
        ((eq 'at-end-of-buffer (car reg-exp-list))
              (concat "\\("
                      (elan-reg-exp-traverse (elan-cadr reg-exp-list))
                      "\\)\\'"))
        ((eq 'at-beginning-of-word (car reg-exp-list))
               (concat "\\<\\("
                      (elan-reg-exp-traverse (elan-cadr reg-exp-list))
                      "\\)"))
        ((eq 'at-end-of-word (car reg-exp-list))
              (concat "\\("
                      (elan-reg-exp-traverse (elan-cadr reg-exp-list))
                      "\\)\\>"))
       ((eq 'any-single-word-constituent-char (car reg-exp-list)) "\\w") 
       ((eq 'any-single-non-word-constituent-char (car reg-exp-list)) "\\W") 
       ((eq 'syntax-code (car reg-exp-list)) 
              (concat "\\s" (elan-reg-exp-syntax-code (elan-cadr reg-exp-list))))
       ((eq 'not-syntax-code (car reg-exp-list)) 
              (concat "\\S" (elan-reg-exp-syntax-code (elan-cadr reg-exp-list))))

       (t (error (concat "Unknown keyword in regular expression: " 
                         (elan-symbol-to-string (car reg-exp-list)))))
  ))

(defun elan-reg-exp-traverse-string (reg-exp-string)
  (elan-reg-exp-escape-special-chars reg-exp-string))

(defun elan-reg-exp-char-set-list-element (x)
  "Handle a symbol within a char set"
  (cond ((eq 'tab x) "\t")
        ((eq 'backspace x ) "\b")
        ((eq 'newline x) "\n")
        ((eq 'vertical-tab x) "\v")
        ((eq 'formfeed x) "\f")
        ((eq 'return x) "\r")
        ((eq 'escape x) "\e")
        ((eq 'space x) " ")
        ((eq 'quote x) "\"")
        (t (error ("Unknown symbol in regular expression")))))

(defun elan-aset-string (str i str1 j)
  "Assign destructively str1 into position i in str. j is a helping, indexing parameter"
  (if (= j (length str1))
      nil  ; result in str
      (progn 
        (aset str i (aref str1 j))
        (elan-aset-string str (+ i 1) str1 (+ j 1)))))

(defun elan-reg-exp-parenthesize (re)
  "Surround the expression RE with parentheses"
  (concat "\\(" re "\\)"))

(defun elan-reg-exp-global-char-set-translate (lst)
  "Make some global changes in LST, which is a list
of char set elements. The changes implement the special
char convention for `]', `-' and `^' in textual regular
expressions. Return the modified list as result."
  (let ((res lst))
    (if (and (= (length lst) 1) (eq (car lst) "^")) 
	(error "Error in char-set: not possible to have single element: '^'"))

    ; force "^" not to be first. 
    (if (equal (car res) "^")
        (setq res (reverse res)))
    (if (equal (car res) "^")
        (error "Weird character set list in char-set"))
 
    ; force "]" to be first:
    (if (member "]" lst)
	(setq res (cons "]" 
			(elan-filter
			 (function
			  (lambda (x)
			    (not (equal x "]"))))
			 lst))))

    ; translate "-" to "--"
    (setq res 
      (mapcar
       (function
         (lambda(x)
           (if (equal x "-") "--" x)))
       res))

    res))

(defun elan-reg-exp-char-set-element(x)
  "Handle a char-set element in a structural regular expression"
  (cond ((symbolp x) (elan-reg-exp-char-set-list-element x))
        ((and (stringp x) (= 1 (length x))) x)
        ((elan-reg-exp-char-range-p x) (concat (elan-reg-exp-char-set-element (elan-cadr x))
                                  "-"
                                  (elan-reg-exp-char-set-element (elan-caddr x))))
        ((elan-reg-exp-char-p x) (char-to-string (elan-cadr x)))
        ((and (stringp x) (equal x "--")) x)
        (t (error "Error in char set"))))

(defun elan-reg-exp-syntax-code (x)
  "Handle symbolic as well as real emacs syntax codes"
  (cond ((symbolp x)
         (cond ((eq x 'whitespace) " ")
               ((eq x 'word-constituent) "w")
               ((eq x 'symbol-constituent) "_")
               ((eq x 'punctuation-character) ".")
               ((eq x 'open-parenthesis) "(")
               ((eq x 'close-parenthesis) ")")
               ((eq x 'close-parenthesis) ")")
               ((eq x 'string-quote) "\"")
               ((eq x 'escape) "\\")
               ((eq x 'character-quote) "/")
               ((eq x 'paired-delimiter) "$")
               ((eq x 'expression-prefix) "'")
               ((eq x 'comment-starter) "<")
               ((eq x 'comment-ender) ">")
               (t (error "Unknown symbolic char-set element"))))
         ((stringp x) x)
         (t (error "Error in syntax code"))))

(defun elan-reg-exp-escape-special-chars (str)
  "In STR, replace the meta characters in regular expressions
with their escaped couterparts. "
  (elan-reg-exp-string-replace-list str '("\\" "$" "^" "." "*" "+" "?" "[" "]" )
                  '("\\\\" "\\$" "\\^" "\\." "\\*" "\\+" "\\?" "\\[" "\\]")))

(defun elan-reg-exp-char-range-p (x)
  "Handle a char-range in a structural regular expression"
  (and (listp x) (eq 'char-range (car x))))

(defun elan-reg-exp-char-p (x)
  (and (consp x) (= 2 (length x)) (eq (car x) 'char)))

(defun elan-reg-exp-string-replace-list (str from-list to-list)
  "Return a copy of STR in which occurences of the single letters strings in FROM-LIST,
which is a list of strings, are substituted by the (geneal) strings in TO-LIST.
From-list and to-list are required to lists of equal lengths. Ealiers substitutions
may be affected by the late substitutions. Thus, the replacements are sequential."
  (let ((strg str)
        (fl from-list)
        (tl to-list))
    (while fl
      (progn
         (setq strg (elan-reg-exp-string-replace strg (car fl) (car tl)))
         (setq fl (cdr fl))
         (setq tl (cdr tl))))
    strg))

(defun elan-reg-exp-string-replace (str single-letter-search-string replace-string)
 "In STR search for occurences of SINGLE-LETTER-SEARCH-STRING and replace
it with occurences of REPLACE-STRING. As the name indicates,
SINGLE-LETTER-SEARCH-STRING is a string of only one letter."
  (let* ((str-list (elan-string-to-list str))
         (str-out-list
           (mapcar
             (function 
               (lambda (ch)
                  (if (eq ch (string-to-char single-letter-search-string))
                      replace-string
                      (char-to-string ch))))
             str-list)))
    (apply (function concat) str-out-list)))

(defun elan-string-to-list (str)
  "Return a list of characters, i.e., integers from the string STR"
 (let ((res nil)
       (strg str))
   (while (> (length strg)  0)
     (progn
        (setq res (cons (string-to-char strg) res))
        (setq strg (substring strg 1))
        ))
   (reverse res)))

